#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Forms;
using HeuristicLab.Clients.OKB.AdminService;
using HeuristicLab.Collections;
using HeuristicLab.Core.Views;
using HeuristicLab.MainForm;
using HeuristicLab.MainForm.WindowsForms;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.Clients.OKB {
  [View("EntityCollection View")]
  [Content(typeof(EntityCollection<>), true)]
  [Content(typeof(IEntityCollection<>), false)]
  public partial class EntityCollectionView<T> : EntityView where T : class, IEntity {
    protected TypeSelectorDialog typeSelectorDialog;

    public new IEntityCollection<T> Content {
      get { return (IEntityCollection<T>)base.Content; }
      set { base.Content = value; }
    }

    public ListView EntitiesListView {
      get { return entitiesListView; }
    }

    public EntityCollectionView() {
      InitializeComponent();
    }

    protected override void DeregisterContentEvents() {
      Content.ItemsAdded -= new CollectionItemsChangedEventHandler<T>(Content_ItemsAdded);
      Content.ItemsRemoved -= new CollectionItemsChangedEventHandler<T>(Content_ItemsRemoved);
      Content.CollectionReset -= new CollectionItemsChangedEventHandler<T>(Content_CollectionReset);
      base.DeregisterContentEvents();
    }
    protected override void RegisterContentEvents() {
      base.RegisterContentEvents();
      Content.ItemsAdded += new CollectionItemsChangedEventHandler<T>(Content_ItemsAdded);
      Content.ItemsRemoved += new CollectionItemsChangedEventHandler<T>(Content_ItemsRemoved);
      Content.CollectionReset += new CollectionItemsChangedEventHandler<T>(Content_CollectionReset);
    }

    protected override void OnContentChanged() {
      base.OnContentChanged();
      while (entitiesListView.Items.Count > 0) RemoveListViewItem(entitiesListView.Items[0]);
      viewHost.Content = null;
      if (Content != null) {
        Caption += " (" + Content.GetType().Name + ")";
        foreach (T entity in Content)
          AddListViewItem(CreateListViewItem(entity));
        AdjustListViewColumnSizes();
        SortItemsListView(SortOrder.Ascending);
      }
    }

    protected override void SetEnabledStateOfControls() {
      base.SetEnabledStateOfControls();
      if (Content == null) {
        addButton.Enabled = false;
        sortAscendingButton.Enabled = false;
        sortDescendingButton.Enabled = false;
        removeButton.Enabled = false;
        entitiesListView.Enabled = false;
        detailsGroupBox.Enabled = false;
      } else {
        addButton.Enabled = !Content.IsReadOnly && !ReadOnly;
        sortAscendingButton.Enabled = entitiesListView.Items.Count > 1;
        sortDescendingButton.Enabled = entitiesListView.Items.Count > 1;
        removeButton.Enabled = !Content.IsReadOnly && !ReadOnly && entitiesListView.SelectedItems.Count > 0;
        entitiesListView.Enabled = true;
        detailsGroupBox.Enabled = entitiesListView.SelectedItems.Count == 1;
      }
    }

    protected virtual T CreateEntity() {
      if (typeSelectorDialog == null) {
        typeSelectorDialog = new TypeSelectorDialog();
        typeSelectorDialog.Caption = "Select Entity";
        typeSelectorDialog.TypeSelector.Caption = "Available Entities";
        typeSelectorDialog.TypeSelector.Configure(typeof(T), false, true);
      }

      if (typeSelectorDialog.ShowDialog(this) == DialogResult.OK) {
        try {
          return (T)typeSelectorDialog.TypeSelector.CreateInstanceOfSelectedType();
        }
        catch (Exception ex) {
          ErrorHandling.ShowErrorDialog(this, ex);
        }
      }
      return null;
    }
    protected virtual ListViewItem CreateListViewItem(T entity) {
      ListViewItem listViewItem = new ListViewItem();
      listViewItem.Text = entity.ToString();
//TODO      listViewItem.ToolTipText = entity.ItemName + ": " + entity.ItemDescription;
      listViewItem.Tag = entity;
      entitiesListView.SmallImageList.Images.Add(entity.EntityImage);
      listViewItem.ImageIndex = entitiesListView.SmallImageList.Images.Count - 1;
      return listViewItem;
    }
    protected virtual void AddListViewItem(ListViewItem listViewItem) {
      entitiesListView.Items.Add(listViewItem);
      ((T)listViewItem.Tag).EntityImageChanged += new EventHandler(Entity_EntityImageChanged);
      ((T)listViewItem.Tag).ToStringChanged += new EventHandler(Entity_ToStringChanged);
      sortAscendingButton.Enabled = entitiesListView.Items.Count > 1;
      sortDescendingButton.Enabled = entitiesListView.Items.Count > 1;
    }
    protected virtual void RemoveListViewItem(ListViewItem listViewItem) {
      ((T)listViewItem.Tag).EntityImageChanged -= new EventHandler(Entity_EntityImageChanged);
      ((T)listViewItem.Tag).ToStringChanged -= new EventHandler(Entity_ToStringChanged);
      listViewItem.Remove();
      foreach (ListViewItem other in entitiesListView.Items)
        if (other.ImageIndex > listViewItem.ImageIndex) other.ImageIndex--;
      entitiesListView.SmallImageList.Images.RemoveAt(listViewItem.ImageIndex);
      sortAscendingButton.Enabled = entitiesListView.Items.Count > 1;
      sortDescendingButton.Enabled = entitiesListView.Items.Count > 1;
    }
    protected virtual void UpdateListViewItemImage(ListViewItem listViewItem) {
      int i = listViewItem.ImageIndex;
      listViewItem.ImageList.Images[i] = ((T)listViewItem.Tag).EntityImage;
      listViewItem.ImageIndex = -1;
      listViewItem.ImageIndex = i;
    }
    protected virtual void UpdateListViewItemText(ListViewItem listViewItem) {
      if (!listViewItem.Text.Equals(listViewItem.Tag.ToString()))
        listViewItem.Text = listViewItem.Tag.ToString();
    }
    protected virtual IEnumerable<ListViewItem> GetListViewItemsForEntity(T entity) {
      foreach (ListViewItem listViewItem in entitiesListView.Items) {
        if (((T)listViewItem.Tag) == entity)
          yield return listViewItem;
      }
    }

    #region ListView Events
    protected virtual void entitiesListView_SelectedIndexChanged(object sender, EventArgs e) {
      removeButton.Enabled = (Content != null) && !Content.IsReadOnly && !ReadOnly && entitiesListView.SelectedItems.Count > 0;
      AdjustListViewColumnSizes();
      if (showDetailsCheckBox.Checked) {
        if (entitiesListView.SelectedItems.Count == 1) {
          T entity = (T)entitiesListView.SelectedItems[0].Tag;
          detailsGroupBox.Enabled = true;
          viewHost.Content = entity;
        } else {
          viewHost.Content = null;
          detailsGroupBox.Enabled = false;
        }
      }
    }
    protected virtual void entitiesListView_KeyDown(object sender, KeyEventArgs e) {
      if (e.KeyCode == Keys.Delete) {
        if ((entitiesListView.SelectedItems.Count > 0) && !Content.IsReadOnly && !ReadOnly) {
          foreach (ListViewItem item in entitiesListView.SelectedItems)
            Content.Remove((T)item.Tag);
        }
      }
    }
    protected virtual void entitiesListView_DoubleClick(object sender, EventArgs e) {
      if (entitiesListView.SelectedItems.Count == 1) {
        T entity = (T)entitiesListView.SelectedItems[0].Tag;
        IContentView view = MainFormManager.MainForm.ShowContent(entity);
        if (view != null) {
          view.ReadOnly = ReadOnly;
          view.Locked = Locked;
        }
      }
    }
    protected virtual void entitiesListView_ItemDrag(object sender, ItemDragEventArgs e) {
      if (!Locked) {
        ListViewItem listViewItem = (ListViewItem)e.Item;
        T entity = (T)listViewItem.Tag;
        DataObject data = new DataObject();
        data.SetData("Type", entity.GetType());
        data.SetData("Value", entity);
        if (Content.IsReadOnly || ReadOnly) {
          DoDragDrop(data, DragDropEffects.Copy | DragDropEffects.Link);
        } else {
          DragDropEffects result = DoDragDrop(data, DragDropEffects.Copy | DragDropEffects.Link | DragDropEffects.Move);
          if ((result & DragDropEffects.Move) == DragDropEffects.Move)
            Content.Remove(entity);
        }
      }
    }
    protected virtual void entitiesListView_DragEnterOver(object sender, DragEventArgs e) {
      e.Effect = DragDropEffects.None;
      Type type = e.Data.GetData("Type") as Type;
      if (!Content.IsReadOnly && !ReadOnly && (type != null) && (typeof(T).IsAssignableFrom(type))) {
        if ((e.KeyState & 32) == 32) e.Effect = DragDropEffects.Link;  // ALT key
        else if ((e.KeyState & 4) == 4) e.Effect = DragDropEffects.Move;  // SHIFT key
        else if ((e.AllowedEffect & DragDropEffects.Copy) == DragDropEffects.Copy) e.Effect = DragDropEffects.Copy;
        else if ((e.AllowedEffect & DragDropEffects.Move) == DragDropEffects.Move) e.Effect = DragDropEffects.Move;
        else if ((e.AllowedEffect & DragDropEffects.Link) == DragDropEffects.Link) e.Effect = DragDropEffects.Link;
      }
    }
    protected virtual void entitiesListView_DragDrop(object sender, DragEventArgs e) {
      if (e.Effect != DragDropEffects.None) {
        T entity = e.Data.GetData("Value") as T;
//TODO        if ((e.Effect & DragDropEffects.Copy) == DragDropEffects.Copy) entity = (T)entity.Clone();
        Content.Add(entity);
      }
    }
    #endregion

    #region Button Events
    protected virtual void addButton_Click(object sender, EventArgs e) {
      T entity = CreateEntity();
      if (entity != null)
        Content.Add(entity);
    }
    protected virtual void sortAscendingButton_Click(object sender, EventArgs e) {
      SortItemsListView(SortOrder.Ascending);
    }
    protected virtual void sortDescendingButton_Click(object sender, EventArgs e) {
      SortItemsListView(SortOrder.Descending);
    }
    protected virtual void removeButton_Click(object sender, EventArgs e) {
      if (entitiesListView.SelectedItems.Count > 0) {
        foreach (ListViewItem item in entitiesListView.SelectedItems)
          Content.Remove((T)item.Tag);
        entitiesListView.SelectedItems.Clear();
      }
    }
    #endregion

    #region CheckBox Events
    protected virtual void showDetailsCheckBox_CheckedChanged(object sender, EventArgs e) {
      if (showDetailsCheckBox.Checked) {
        splitContainer.Panel2Collapsed = false;
        detailsGroupBox.Enabled = entitiesListView.SelectedItems.Count == 1;
        viewHost.Content = entitiesListView.SelectedItems.Count == 1 ? (T)entitiesListView.SelectedItems[0].Tag : null;
      } else {
        splitContainer.Panel2Collapsed = true;
        viewHost.Content = null;
      }
    }
    #endregion

    #region Content Events
    protected virtual void Content_ItemsAdded(object sender, CollectionItemsChangedEventArgs<T> e) {
      if (InvokeRequired)
        Invoke(new CollectionItemsChangedEventHandler<T>(Content_ItemsAdded), sender, e);
      else {
        foreach (T item in e.Items)
          AddListViewItem(CreateListViewItem(item));
        AdjustListViewColumnSizes();
      }

    }
    protected virtual void Content_ItemsRemoved(object sender, CollectionItemsChangedEventArgs<T> e) {
      if (InvokeRequired)
        Invoke(new CollectionItemsChangedEventHandler<T>(Content_ItemsRemoved), sender, e);
      else {
        foreach (T item in e.Items) {
          // remove only the first matching ListViewItem, because the IEntity could be contained multiple times in the EntityCollection
          ListViewItem listviewItem = GetListViewItemsForEntity(item).FirstOrDefault();
          if (listviewItem != null) RemoveListViewItem(listviewItem);
        }
      }
    }
    protected virtual void Content_CollectionReset(object sender, CollectionItemsChangedEventArgs<T> e) {
      if (InvokeRequired)
        Invoke(new CollectionItemsChangedEventHandler<T>(Content_CollectionReset), sender, e);
      else {
        foreach (T item in e.OldItems) {
          // remove only the first matching ListViewItem, because the IEntity could be contained multiple times in the EntityCollection
          ListViewItem listviewItem = GetListViewItemsForEntity(item).FirstOrDefault();
          if (listviewItem != null) RemoveListViewItem(listviewItem);
        }
        foreach (T item in e.Items)
          AddListViewItem(CreateListViewItem(item));
      }
    }
    #endregion

    #region Entity Events
    protected virtual void Entity_EntityImageChanged(object sender, EventArgs e) {
      if (InvokeRequired)
        Invoke(new EventHandler(Entity_EntityImageChanged), sender, e);
      else {
        T item = (T)sender;
        foreach (ListViewItem listViewItem in GetListViewItemsForEntity(item))
          UpdateListViewItemImage(listViewItem);
      }
    }
    protected virtual void Entity_ToStringChanged(object sender, EventArgs e) {
      if (InvokeRequired)
        Invoke(new EventHandler(Entity_ToStringChanged), sender, e);
      else {
        T item = (T)sender;
        foreach (ListViewItem listViewItem in GetListViewItemsForEntity(item))
          UpdateListViewItemText(listViewItem);
      }
    }
    #endregion

    #region Helpers
    protected virtual void SortItemsListView(SortOrder sortOrder) {
      entitiesListView.Sorting = SortOrder.None;
      entitiesListView.Sorting = sortOrder;
      entitiesListView.Sorting = SortOrder.None;
    }
    protected virtual void AdjustListViewColumnSizes() {
      if (entitiesListView.Items.Count > 0) {
        for (int i = 0; i < entitiesListView.Columns.Count; i++)
          entitiesListView.Columns[i].AutoResize(ColumnHeaderAutoResizeStyle.ColumnContent);
      }
    }
    #endregion
  }
}
